# -*- coding: utf-8 -*-
"""
This module contains objects for handling indexing in Dynochemy.

This is a framework where an index can be defined that either 'adds' or 'removes' a entity and 
then emits sets of operations that should be performed.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""



class Index(object):
    def add(self, entity):
        pass
    def remove(self, entity):
        pass




# -*- coding: utf-8 -*-
"""
This module contains constants that are needed in multiple modules.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""

MAX_BATCH_WRITE_ITEMS = 25
MAX_BATCH_READ_ITEMS = 100




# -*- coding: utf-8 -*-
"""
Instrumented Dynochemy Database
We use 'BlueOx' to provide useful stats and reporting for requests to dynochemy

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""
import logging
import time
import json

from . import db
from . import sql

# Don't fail if blueox isn't supported.
try:
    import blueox
except ImportError:
    blueox = None


class OxClient(object):
    def __init__(self, client):
        self.real_client = client

    def register_table(self, *args, **kwargs):
        return self.real_client.register_table(*args, **kwargs)

    def make_request(self, command, body=None, callback=None):
        start_time = time.time()

        def wrapped_callback(*args, **kwargs):
            with blueox.Context('.dynochemy') as ctx:
                # Rewrite our start time, this is async, remember
                ctx.start_time = start_time
                ctx.set('command', command)

                resp = args[0]
                err = kwargs.get('error')

                if 'Responses' in resp:
                    for table_name, table_resp in resp['Responses'].iteritems():
                        if 'ConsumedCapacityUnits' in table_resp:
                            ctx.set('tables.%s.consumed_capacity' % table_name, table_resp['ConsumedCapacityUnits'])
                elif 'ConsumedCapacityUnits' in resp:
                    req = json.loads(body)
                    ctx.set('tables.%s.consumed_capacity' % req.get('TableName', ''), resp['ConsumedCapacityUnits'])

                if 'Count' in resp:
                    ctx.set('count', resp['Count'])

                if err:
                    ctx.set('error', str(err))

            if callback:
                callback(*args, **kwargs)

        return self.real_client.make_request(command, body=body, callback=wrapped_callback)


class OxDB(db.DB):
    def __init__(self, access_key, access_secret, ioloop=None):
        super(OxDB, self).__init__(access_key, access_secret, ioloop=ioloop)

        # Wrap the client
        self.client = OxClient(self.client)


class OxSQLDB(sql.SQLDB):
    def __init__(self, engine):
        super(OxSQLDB, self).__init__(engine)

        # Wrap the client
        self.client = OxClient(self.client)



# -*- coding: utf-8 -*-

"""
This module contains the set of Dynochemy's exceptions

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.

"""
import json


class Error(Exception):
    """This is an ambiguous error that occured."""
    pass

class SyncUnallowedError(Error): pass

class DuplicateBatchItemError(Error): pass

class IncompleteSolventError(Error): pass

class ExceededBatchRequestsError(Error): pass

class ItemNotFoundError(Error): pass

class DynamoDBError(Error): pass

class ProvisionedThroughputError(DynamoDBError): pass

class UnprocessedItemError(DynamoDBError): pass



def parse_error(raw_error):
    """Parse the error we get out of Boto into something we can code around"""
    if isinstance(raw_error, Error):
        return raw_error

    if 'ProvisionedThroughputExceededException' in raw_error.error_code:
        return ProvisionedThroughputError(raw_error.error_message)
    else:
        return DynamoDBError(raw_error.error_message, raw_error.error_code)


__all__ = ["Error", "SyncUnallowedError", "DuplicateBatchItemError", "DynamoDBError", "ProvisionedThroughputError", "ItemNotFoundError"]



# -*- coding: utf-8 -*-
"""
This module contains classes for abstract Dynochemy operations.

There are a few levels of usage here:

    * Primitive Operations (Get, Put, Delete, Update)
    * Combined Operations (single requests with multiple operations in them: BatchRead, BatchWrite)
    * Operation sequences (Batches with too many operations to execute at once, Queries and Scans)

Note that any instance of 'Operation' can be run individually, but the real
power come from allowing a higher power to manage the execution (see solvent)

Another important object is the OperationResult, which provides a way for
results to be store the result of a combined operation, but keyed by primitive
operation. To use an operation directly, an OperationResult object must be provided.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""
import functools
import itertools
import copy
import logging

from . import errors
from . import defer
from . import utils
from . import constants

log = logging.getLogger(__name__)

def operation_clone_defer(df):
    """Defer's come from the db with extra parameters. We want to clean those up."""
    new_df = defer.Defer()
    if df.error:
        # Copy over the error. Might be more efficient ways of doing this.
        try:
            df.rethrow()
        except Exception:
            new_df.exception()
        else:
            assert False, "Should have an exception"
    else:
        new_df.callback(df.args[0])

    return new_df

class Operation(object):
    """(Abstract)Base class for all operations.

    """
    def __init__(self, noindex=False):
        self.noindex = noindex

    def run_defer(self, op_results):
        raise NotImplementedError

    @property
    def unique_key(self):
        raise NotImplementedError

    def __eq__(self, other):
        return self.unique_key == other.unique_key

    def __hash__(self):
        return hash(self.unique_key)

    def run(self, op_results):
        # We require passing in an OperationResult object which might seem
        # weird.  This is one reason why Operation should really be considered
        # an internal API... or at the very least, should probalby never be
        # 'run()' directly.  In an earlier design, we provided the
        # OperationResult for each operation, and then the Solvent() (the more
        # common interface) took care of combining all these results together.
        # However, this proved to be really hard to understand and debug.

        # So in the interest of our sanity, an operation is provied the object
        # it will deliver it's results through.

        df = self.run_defer(op_results)

        return df()

    def run_async(self, op_results, callback=None):
        df = self.run_defer(op_results)

        def handle_result(cb):
            if callback is not None:
                callback(*cb.result)

        df.add_callback(handle_result)

    def have_result(self, op_results, db_df, ignore_capacity=False):
        """Called when a result for this operation is available.

        Args -
          op_result: OperationResult object we have been storing results into for this operation.
          op_cb: The defer making the callback
          ignore_capacity: Set to true if this result should not record capacity. This would be the case if this operation wasn't run individually.

        This gives an operation the chance to intercept the processing of results, and potentially queue 'next operations'.
        """

        # Operations come from the database with args and keywords that include
        # capacity. We want to strip that out so our results can be used nice
        # and cleanly with a single return value
        assert db_df.done
        op_df = operation_clone_defer(db_df)
        
        if ignore_capacity or db_df.kwargs is None:
            op_results.record_result(self, op_df)
        else:
            op_results.record_result(self, 
                             op_df, 
                             read_capacity=db_df.kwargs.get('read_capacity'), 
                             write_capacity=db_df.kwargs.get('write_capacity'))


def combine_dicts(left_dict, right_dict, combiner=None):
    """Utility function for combining two dictionaries (union) with a user specified 'combiner' function

    Note that the default combiner just takes the value in right first, followed by left if they are truthy
    """
    if combiner is None:
        def combiner(l, r):
            return r or l

    if not any((left_dict, right_dict)):
        return {}

    if not left_dict and right_dict:
        # Swap arguments so we have some value
        s_dict = left_dict
        left_dict = right_dict
        right_dict = {}

    right_dict = right_dict or {}

    out_dict = left_dict.copy()
    for k, v in right_dict.iteritems():
        out_dict[k] = combiner(left_dict.get(k), v)
    return out_dict


class UpdateOperation(Operation):
    def __init__(self, table, key, add=None, put=None, delete=None, **kwargs):
        super(UpdateOperation, self).__init__(**kwargs)
        self.table = table
        self.key = key
        self.add = add
        self.put = put
        self.delete = delete

    def combine_updates(self, update_op):
        """Combine two UpdateOperations assuming they are for the same table/key combination"""
        assert self.table == update_op.table
        assert self.key == update_op.key

        new_update = UpdateOperation(self.table, self.key)

        def add_combiner(l, r):
            if l and r:
                return l + r
            else:
                return l or r

        new_update.add = combine_dicts(self.add, update_op.add, combiner=add_combiner)
        new_update.put = combine_dicts(self.put, update_op.put)
        new_update.delete = combine_dicts(self.delete, update_op.delete)

        return new_update

    def run_defer(self, op_results):
        update_df = getattr(op_results.db, self.table.__name__).update_defer(self.key, add=self.add, put=self.put, delete=self.delete, return_value="ALL_OLD")
        update_df.add_callback(functools.partial(self.have_result, op_results))

        return update_df

    def result(self, old_result):
        """Return what we think the result of this operation will be, given the old entity.

        This is useful because we typically retrieve the old value from the db
        so we can properly understand the changes that took place.  We can't
        just get the new value from the db, because sometimes views need to
        know what changed (for example, counters)
        """
        new_item = copy.copy(old_result)

        if self.put:
            for attribute, value in self.put.iteritems():
                new_item[attribute] = value

        if self.add:
            for attribute, value in self.add.iteritems():
                if attribute in new_item:
                    if isinstance(new_item[attribute], (int, float)):
                        new_item[attribute] += value
                    elif isinstance(new_item[attribute], list):
                        if hasattr(value, '__iter__'):
                            new_item[attribute] += [v for v in value]
                        else:
                            new_item[attribute].append(value)
                    else:
                        new_item[attribute].append(value)
                else:
                    new_item[attribute] = value

        if self.delete:
            for attribute, value in self.delete.iteritems():
                del new_item[attribute]

        return new_item

    @property
    def unique_key(self):
        return ('UPDATE', self.table.name, self.key)

    def __repr__(self):
        return "<UpdateOperation %s:%r>" % (self.table.name, self.key,)


class _WriteBatchableMixin(object):
    """Mixing for operations that can be put in a batch write"""
    def add_to_batch(self, batch):
        raise NotImplementedError


class _ReadBatchableMixin(object):
    """Mixing for operations that can be put in a batch read"""
    def add_to_batch(self, batch):
        raise NotImplementedError


class BatchOperation(Operation):
    # BatchOperations have a slightly more complex design as they need to track sub-operations.
    # We have a few requiremets:
    #   1. Report results on an individual op basis (a defer for each sub-op)
    #   2. Generate follow on batches for sub-ops that didn't get included due to being over capacity

    __slots__ = ["ops"]
    def __init__(self, ops=None, **kwargs):
        super(BatchOperation, self).__init__(**kwargs)
        self.ops = set()

        if ops:
            for op in ops:
                self.add(op)

    def add(self, op):
        self.ops.add(op)

    def __len__(self):
        return len(self.ops)

    def __iter__(self):
        for op in self.ops:
            yield op


class BatchWriteOperation(BatchOperation):
    def add(self, op):
        if not isinstance(op, _WriteBatchableMixin):
            raise ValueError(op)

        if isinstance(op, BatchWriteOperation):
            self.ops.update(op.ops)
        else:
            self.ops.add(op)

    def run_defer(self, op_results):
        if not self.ops:
            df = defer.Defer()
            df.done = True
            return df

        def handle_op_result(op, cb):
            op.have_result(op_results, cb, ignore_capacity=True)

        def handle_batch_result(cb):
            # Note that we are not recording the full result for this op.
            # This means that in the case of an error, there will be errors for each sub-op, but not
            # one for the overall batch.
            # We do however need to record the capacity used, which we can only do in aggregate.
            if 'write_capacity' in cb.kwargs:
                op_results.update_write_capacity(cb.kwargs['write_capacity'])

        # We might have more operations than we can put in single batch, so prepare ourselves.
        all_ops = list(self.ops)
        batch = op_results.db.batch_write()

        while all_ops:
            op = all_ops.pop()
            try:
                op_df = op.add_to_batch(batch)
            except errors.ExceededBatchRequestsError:
                # Too many requests, put it back
                all_ops.append(op)
                break
            else:
                # To collect our results per sub-operation, we'll need to add a
                # hook to record our specific part of the response for each
                # sub-op
                op_df.add_callback(functools.partial(handle_op_result, op))

        batch_df = batch.defer()
        batch_df.add_callback(handle_batch_result)

        # If there are more batches left, leave them for the next round.
        if all_ops:
            op_results.next_ops += all_ops

        return batch_df

    def __repr__(self):
        return "<BatchWriteOperation:%d>" % (len(self),)


class BatchReadOperation(BatchOperation):
    def add(self, op):
        if not isinstance(op, _ReadBatchableMixin):
            raise ValueError(op)
        if isinstance(op, BatchReadOperation):
            self.ops.update(op.ops)
        else:
            self.ops.add(op)

    def run_defer(self, op_results):
        # The implementation here is very similiar to BatchRead, so you should
        # probalby reference that for more explanation.
        if not self.ops:
            df = defer.Defer()
            df.done = True
            return df

        def handle_op_result(op, cb):
            op.have_result(op_results, cb)

        def handle_batch_result(cb):
            op_results.update_read_capacity(cb.kwargs.get('read_capacity', {}))

        all_ops = list(self.ops)

        batch = op_results.db.batch_read()
        while all_ops:
            op = all_ops.pop()
            try:
                op_df = op.add_to_batch(batch)
            except errors.ExceededBatchRequestsError:
                all_ops.append(op)
                break
            else:
                op_df.add_callback(functools.partial(handle_op_result, op))

        if all_ops:
            op_results.next_ops += all_ops

        batch_df = batch.defer()
        batch_df.add_callback(handle_batch_result)

        return batch_df

    def __repr__(self):
        return "<BatchReadOperation:%d>" % (len(self),)


class PutOperation(Operation, _WriteBatchableMixin):
    def __init__(self, table, entity, **kwargs):
        super(PutOperation, self).__init__(**kwargs)
        self.table = table
        self.entity = entity

    def run_defer(self, op_results):
        op_df = getattr(op_results.db, self.table.__name__).put_defer(self.entity)
        op_df.add_callback(functools.partial(self.have_result, op_results))

        return op_df

    def add_to_batch(self, batch):
        return getattr(batch, self.table.__name__).put(self.entity)

    @property
    def key(self):
        return tuple([self.entity[k] for k in self.table(None).key_spec])

    @property
    def unique_key(self):
        return ('PUT', self.table.name, self.key)

    def __repr__(self):
        return "<PutOperation %s:%r>" % (self.table.name, self.key,)


class DeleteOperation(Operation, _WriteBatchableMixin):
    def __init__(self, table, key, **kwargs):
        super(DeleteOperation, self).__init__(**kwargs)
        self.table = table
        self.key = key

    def run_defer(self, op_results):
        op_df = getattr(op_results.db, self.table.__name__).delete_defer(self.key)
        op_df.add_callback(functools.partial(self.have_result, op_results))

        return op_df

    def add_to_batch(self, batch):
        return getattr(batch, self.table.__name__).delete(self.key)

    @property
    def unique_key(self):
        # Assumes updates for the same key have been combined together
        return ('DELETE', self.table.name, self.key)

    def __repr__(self):
        return "<DeleteOperation %s:%r>" % (self.table.name, self.key,)


class GetOperation(Operation, _ReadBatchableMixin):
    def __init__(self, table, key, **kwargs):
        super(GetOperation, self).__init__(**kwargs)
        self.table = table
        self.key = key

    def run_defer(self, op_results):
        op_df = getattr(op_results.db, self.table.__name__).get_defer(self.key)
        op_df.add_callback(functools.partial(self.have_result, op_results))

        return op_df

    def add_to_batch(self, batch):
        return getattr(batch, self.table.__name__).get(self.key)

    @property
    def unique_key(self):
        # Assumes updates for the same key have been combined together
        return ('GET', self.table.name, self.key)

    def __repr__(self):
        return "<GetOperation %s:%r>" % (self.table.name, self.key,)


class GetAndDeleteOperation(GetOperation):
    """Operation that does a Get, then a Delete

    This is a replacement for doing just a straight Delete. First it does a Get, then it schedules a Delete to go afterwards.
    This is useful in solvents because for views, to remove an entity, you really need to know what the entity is.
    """
    def have_result(self, op_results, op_cb):
        super(GetAndDeleteOperation, self).have_result(op_results, op_cb)

        try:
            entity = op_results[self]
        except Exception:
            entity = None

        if entity:
            op_results.next_ops.append(DeleteOperation(self.table, self.key))

    def __repr__(self):
        return "<GetAndDeleteOperation %s:%r>" % (self.table.name, self.key,)


class QueryOperation(Operation):
    """Combined query operation that runs multiple sub-queries until retieving all the requested results.
   
    What this really means, is that doing multiple individual query requests is
    handled directly by the Operation, rather than by a higher-level solvent,
    which I would prefer. 
    
    This issue is that the results of each
    QuerySegmentOperation need to be combined together intelligently but there
    isn't currently a real clean way for a solvent to sort that out for us.

    This big problem is (besides how complex it is) how we handle errors and
    retries. We could leave provisioning error issues to our 
    solvent, but that means we'll re-run the entire query operation, not just a segment.
    """
    def __init__(self, table, key, args=None, **kwargs):
        super(QueryOperation, self).__init__(**kwargs)
        self.table = table
        self.hash_key = key
        self.total_limit = None
        self.args = args or {}

    def __copy__(self):
        op = self.__class__(self.table, self.hash_key, args=copy.copy(self.args))
        op.total_limit = self.total_limit
        return op

    def range(self, start=None, end=None):
        self.args['range'] = (start, end)
        return self

    def reverse(self, reverse=True):
        self.args['reverse'] = reverse
        return self

    def limit(self, limit):
        self.total_limit = int(limit)
        self.args['limit'] = int(limit)
        return self

    def last_seen(self, range_id):
        self.args['last_seen'] = range_id
        return self

    @property
    def unique_key(self):
        # We are not including the 'last_seen' argument in our hash because we
        # want all the different segments of our query to combine together.
        return ('QUERY', self.table.name, self.hash_key, tuple((k, v) for k, v in self.args.iteritems() if k not in ('last_seen', 'limit')), self.total_limit)

    def have_result(self, op_results, db_df, **kwargs):
        # Query result handling is special.
        # Queries can happen in multiple operations.
        # We need to potentially update our resulting QueryResult object with new entries in place.
        # Also, we might need to queue up our operation again with different bounds if there are more results available.


        # HEREIAM: Difficult with error handling with these new defers. We seem to record a failure in a multi-step query by keeping
        # the old result object, but updating with a new error. Later on, when checking for errors to retry, we ignore the results, just 
        # looking for errors.

        op_df = operation_clone_defer(db_df)

        try:
            new_query_result = op_df.result
            new_err = None
        except Exception, e:
            new_err = e

        if self in op_results and op_results.results[self].args:
            # We already have part of this query, update our results in place.
            old_query_result = op_results.raw_result(self)

            if new_err:
                # Mark us as an error
                op_results.set_raw_error(self, new_err)
            else:
                op_results.set_raw_result(self, old_query_result.combine(new_query_result))
                if 'read_capacity' in db_df.kwargs:
                    op_results.update_read_capacity(db_df.kwargs['read_capacity'])
        else:
            super(QueryOperation, self).have_result(op_results, db_df)

        if self.total_limit:
            # Let's make sure we never return more entities than our limit
            full_res = op_results[self]
            if full_res:
                assert len(full_res) <= self.total_limit

        if new_err:
            return
        elif new_query_result:
            # What do we do next? If we've met our limit, we should stop.
            # Otherwise we might need to queue up the next page of results.
            current_count = len(new_query_result)
            if self.args.get('limit') and current_count >= int(self.args['limit']):
                return
            elif new_query_result.has_next:
                last_key = utils.parse_key(new_query_result.result_data['LastEvaluatedKey'])
                next_op = copy.copy(self)
                next_op.last_seen(last_key[1])

                if self.args.get('limit'):
                    next_op.args['limit'] = self.args['limit'] - current_count
                    assert next_op.args['limit'] > 0

                op_results.next_ops.append(next_op)

    def run_defer(self, op_results):
        query = getattr(op_results.db, self.table.__name__).query(self.hash_key)
        for param, arg in self.args.iteritems():
            if param == 'range':
                query = query.range(*arg)
            else:
                query = getattr(query, param)(arg)

        op_df = query.defer()
        op_df.add_callback(functools.partial(self.have_result, op_results))
        return op_df

    def __repr__(self):
        return "<QueryOperation %s:%r>" % (self.table.name, self.hash_key,)


class QueryAndDeleteOperation(QueryOperation):
    """An operation that does a query and deletes all the results.

    You can provide an additional filter function (filter_func) that will be
    applied as a filter to all the results from the query.  If the filter
    returns False, the entity will not be deleted.
    """
    def __init__(self, table, key, filter_func=None, **kwargs):
        super(QueryAndDeleteOperation, self).__init__(table, key, **kwargs)
        self.filter_func = filter_func

    def __copy__(self):
        op = self.__class__(self.table, self.hash_key, args=copy.copy(self.args), filter_func=self.filter_func)
        return op

    def have_result(self, op_results, op_cb, **kwargs):
        super(QueryAndDeleteOperation, self).have_result(op_results, op_cb, **kwargs)

        filter_func = self.filter_func or (lambda e: True)

        try:
            query_result = op_cb.result
        except Exception:
            # Nothing to be done
            pass
        else:
            for res in filter(filter_func, query_result):
                op_results.next_ops.append(DeleteOperation(self.table, (res[self.table.hash_key], res[self.table.range_key])))

    def __repr__(self):
        return "<QueryAndDeleteOperation %s:%r>" % (self.table.name, self.hash_key,)


class OperationResultDefer(defer.Defer):
    """Special defer that is bound to a result before it's completed.

    The only 'result' that can come from the normal 'callback()' flow is in the case of errors.
    """
    def __init__(self, op_result, io_loop):
        super(OperationResultDefer, self).__init__(io_loop)
        self.op_result = op_result

    @property
    def result(self):
        # Just a sanity check, there shouldn't be any real results
        args, kwargs = super(OperationResultDefer, self).result
        assert not args, args
        assert not kwargs, kwargs

        return self.op_result


class OperationResult(object):
    def __init__(self, db):
        self.db = db
        self.results = {}
        self.callbacks = []

        self.read_capacity = {}
        self.write_capacity = {}

        self.next_ops = []
        self.error_attempts = 0

    def record_result(self, op, df, read_capacity=None, write_capacity=None):
        self.results[op] = df

        if read_capacity:
            self.update_read_capacity(read_capacity)

        if write_capacity:
            self.update_write_capacity(write_capacity)

        # Inform our result callbacks
        [func(op) for func in self.callbacks]

    def rethrow(self):
        for df in self.results.values():
            df.rethrow()

    def add_callback(self, func):
        """Add a function to call when we get a result.

        Argument will be an operation instance.
        Generally the caller should know the result object and can look up the result itself.
        """
        self.callbacks.append(func)

    def update(self, other_result):
        assert self.db == other_result.db
        self.results.update(other_result.results)

        self.update_read_capacity(other_result.read_capacity)
        self.update_write_capacity(other_result.write_capacity)

        self.next_ops += other_result.next_ops

    def update_read_capacity(self, read_capacity):
        for name, value in read_capacity.iteritems():
            self.read_capacity.setdefault(name, 0.0)
            self.read_capacity[name] += value

    def update_write_capacity(self, write_capacity):
        for name, value in write_capacity.iteritems():
            self.write_capacity.setdefault(name, 0.0)
            self.write_capacity[name] += value

    # These are some rarely need methods to poke at the underlying operation df.
    def set_raw_result(self, op, result):
        self.results[op].args = [result]

    def raw_result(self, op):
        return self.results[op].args[0]

    def set_raw_error(self, op, ex):
        self.results[op].exception(ex=ex)

    def raw_error(self, op):
        return self.results[op].error

    def iteritems(self):
        for op, df in self.results.iteritems():
            yield (op, df.result)

    def __iter__(self):
        for op in self.results.keys():
            yield op

    def __getitem__(self, key):
        return self.results[key].result

    def __delitem__(self, key):
        del self.results[key]

    def __repr__(self):
        return repr(self.results)


__all__ = ["GetOperation", "PutOperation", "DeleteOperation", "UpdateOperation", "QueryOperation", "BatchReadOperation", "BatchWriteOperation", "OperationResult"]



# -*- coding: utf-8 -*-
"""
SQLite emulation of DynamoDB
This is useful for local development and testing.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""
import json
import collections
import pprint

import sqlalchemy
from sqlalchemy import sql, Table, Column, String, LargeBinary

from dynochemy import db
from dynochemy import errors
from dynochemy import utils

DEFAULT_LIMIT = 100

def create_db(db_name, metadata, has_range_key=True):
    if has_range_key:
        dynamo_tbl = Table(db_name, metadata,
                           Column('hash_key', String, primary_key=True),
                           Column('range_key', String, primary_key=True),
                           Column('content', LargeBinary),
                        )
    else:
        dynamo_tbl = Table(db_name, metadata,
                           Column('hash_key', String, primary_key=True),
                           Column('content', LargeBinary),
                        )

    return dynamo_tbl


class CommandError(errors.Error): pass


def compare(operator, lval, rval):
    if operator == 'EQ':
        return lval == rval
    elif operator == 'GT':
        return lval > rval
    elif operator == 'LT':
        return lval < rval
    else:
        raise NotImplementedError(operator)


class SQLClient(object):
    def __init__(self, engine):
        self.engine = engine
        self.connection = None

        self.tables = {}
        self.metadata = sqlalchemy.MetaData(bind=engine)

        # During an operation, we keep track of how much capacity is being used per table
        self.reset_op_capacity()

    def reset_op_capacity(self):
        self.table_read_op_capacity = collections.defaultdict(int)
        self.table_write_op_capacity = collections.defaultdict(int)

    def register_table(self, table_cls):
        try:
            sql_table = Table(table_cls.name, self.metadata, autoload=True, autoload_with=self.engine)
        except sqlalchemy.exc.NoSuchTableError:
            sql_table = create_db(table_cls.name, self.metadata, has_range_key=bool(table_cls.range_key))
            sql_table.create(self.engine)

        if table_cls.range_key:
            key_spec = (table_cls.hash_key, table_cls.range_key)
        else:
            key_spec = (table_cls.hash_key,)

        read_counter = utils.ResourceCounter()
        if table_cls.read_capacity is not None:
            read_counter.set_limit(1, table_cls.read_capacity)
            read_counter.set_limit(5, table_cls.read_capacity * 0.80)

        write_counter = utils.ResourceCounter()
        if table_cls.write_capacity is not None:
            write_counter.set_limit(1, table_cls.write_capacity)
            write_counter.set_limit(5, table_cls.write_capacity * 0.80)

        self.tables[table_cls.name] = {
            'sql_table': sql_table,
            'table_def': table_cls,
            'key_spec': key_spec,
            'read_counter': read_counter,
            'write_counter': write_counter,
        }

    def _get_item(self, table_name, key):
        table_spec = self.tables[table_name]
        sql_table = table_spec['sql_table']
        table_def = table_spec['table_def']

        if table_def.range_key:
            expr = sql.and_(sql_table.c.hash_key == utils.parse_value(key['HashKeyElement']), 
                              sql_table.c.range_key == utils.parse_value(key['RangeKeyElement']))
        else:
            expr = sql_table.c.hash_key == utils.parse_value(key['HashKeyElement'])

        q = sql.select([sql_table], expr)
        try:
            res = list(self.engine.execute(q))[0]
        except IndexError:
            return None
        else:
            capacity_size = (len(res[sql_table.c.content]) / 1024) + 1
            self.table_read_op_capacity[table_name] += capacity_size
            self.tables[table_name]['read_counter'].record(capacity_size)
            return json.loads(res[sql_table.c.content])

    def do_getitem(self, args):
        item = self._get_item(args['TableName'], args['Key'])

        if item:
            out = {
                'Item': item
            }
        else:
            out = {}

        return out

    def _put_item(self, table_name, item_data):
        table_spec = self.tables[table_name]
        sql_table = table_spec['sql_table']
        table_def = table_spec['table_def']
        key_spec = table_spec['key_spec']

        item = utils.parse_item(item_data)
        hash_key = item[key_spec[0]]

        item_json = json.dumps(item_data)

        capacity_size = (len(item_json) / 1024) + 1
        self.table_write_op_capacity[table_name] += capacity_size
        self.tables[table_name]['write_counter'].record(capacity_size)

        if table_def.range_key:
            range_key = item[key_spec[1]]
            del_q = sql_table.delete().where(sql.and_(sql_table.c.hash_key==hash_key, sql_table.c.range_key==range_key))
            ins = sql_table.insert().values(hash_key=hash_key, range_key=range_key, content=item_json)
        else:
            del_q = sql_table.delete().where(sql_table.c.hash_key==hash_key)
            ins = sql_table.insert().values(hash_key=hash_key, content=item_json)

        cnxn = self.connection or self.engine

        try:
            cnxn.execute(ins)
        except sqlalchemy.exc.IntegrityError:
            # Try again
            txn = None
            if cnxn == self.engine:
                cnxn = self.engine.connect()
                txn = cnxn.begin()

            cnxn.execute(del_q)
            cnxn.execute(ins)

            if txn:
                txn.commit()
        
        return capacity_size

    def do_putitem(self, args):
        item_data = args['Item']
        if 'Expected' in args:
            raise NotImplementedError

        self._put_item(args['TableName'], item_data)

    def do_deleteitem(self, args):
        table_spec = self.tables[args['TableName']]
        sql_table = table_spec['sql_table']
        table_def = table_spec['table_def']

        if table_def.range_key:
            expr = sql.and_(sql_table.c.hash_key == utils.parse_value(args['Key']['HashKeyElement']), 
                              sql_table.c.range_key == utils.parse_value(args['Key']['RangeKeyElement']))
        else:
            expr = sql_table.c.hash_key == utils.parse_value(args['Key']['HashKeyElement'])

        item = self.do_getitem(args)

        # This one is kinda weird, we're basically gonna transfer whatever we
        # racked up in reads to the write column.
        op_capacity = self.table_read_op_capacity[args['TableName']]
        self.table_read_op_capacity[args['TableName']] = 0
        self.tables[args['TableName']]['read_counter'].record(op_capacity * -1)

        self.table_write_op_capacity[args['TableName']] += op_capacity
        self.tables[args['TableName']]['write_counter'].record(op_capacity)

        del_q = sql_table.delete().where(expr)
        res = self.engine.execute(del_q)

        out = {
            'Attributes': item['Item']
        }

        return out

    def do_updateitem(self, args):
        table_spec = self.tables[args['TableName']]
        sql_table = table_spec['sql_table']
        key_spec = table_spec['key_spec']
        table_def = table_spec['table_def']

        if 'Expected' in args:
            raise NotImplementedError

        if args.get('ReturnValues') not in (None, 'ALL_OLD', 'ALL_NEW'):
            raise NotImplementedError

        if table_def.range_key:
            key = (utils.parse_value(args['Key']['HashKeyElement']), 
                   utils.parse_value(args['Key']['RangeKeyElement']))
            expr = sql.and_(sql_table.c.hash_key == key[0],
                              sql_table.c.range_key == key[1])
        else:
            key = (utils.parse_value(args['Key']['HashKeyElement']),)
            expr = sql_table.c.hash_key == key[0]

        # Update is one of our few transactionally important operations.  By
        # setting self.connection, our callees should use that rather than the
        # connectionless self.engine method, allowing us to control the
        # transaction directly.
        self.connection = self.engine.connect()
        txn = self.connection.begin()

        q = sql.select([sql_table], expr)
        res = list(self.connection.execute(q))
        if res:
            item = json.loads(res[0][sql_table.c.content])
        else:
            item = {}
            item.update(utils.format_key(key_spec, key))

        real_item = utils.parse_item(item)

        # Apply our updates
        for attribute, value_update in args['AttributeUpdates'].iteritems():
            if value_update['Action'] == "ADD":
                if attribute in real_item:
                    if isinstance(real_item[attribute], (int,float,list)):
                        real_item[attribute] += utils.parse_value(value_update['Value'])
                    elif isinstance(real_item[attribute], list):
                        if hasattr(value_update['Value'], '__iter__'):
                            real_item[attribute] += [utils.parse_value(v) for v in value_update['Value']]
                        else:
                            real_item[attribute].append(utils.parse_value(value_update['Value']))
                    else:
                        real_item[attribute].append(utils.parse_value(value_update['Value']))
                else:
                    real_item[attribute] = utils.parse_value(value_update['Value'])

            elif value_update['Action'] == "PUT":
                real_item[attribute] = utils.parse_value(value_update['Value'])
            elif value_update['Action'] == "DELETE":
                if attribute in real_item:
                    del real_item[attribute]
            else:
                raise ValueError(value_update['Action'])
        
        # write to the db
        self._put_item(args['TableName'], utils.format_item(real_item))

        txn.commit()
        self.connection = None

        if args.get('ReturnValues', 'NONE') == 'NONE':
            return {}
        elif args['ReturnValues'] == 'ALL_NEW':
            return {'Attributes': utils.format_item(real_item)}
        elif args['ReturnValues'] == 'ALL_OLD':
            return {'Attributes': item}

    def do_batchwriteitem(self, args):
        out = {}
        for table_name, requests in args['RequestItems'].iteritems():
            for request in requests:
                req_type = request.keys()[0]
                args = request[req_type]
                args['TableName'] = table_name

                if req_type == "PutRequest":
                    if not self.tables[table_name]['write_counter'].check():
                        out.setdefault('UnprocessedItems', {})
                        out['UnprocessedItems'].setdefault(table_name, [])
                        out['UnprocessedItems'][table_name].append(request)
                    else:
                        self._put_item(table_name, args['Item'])

                elif req_type == "DeleteRequest":
                    # TODO: write capacity checks
                    self.do_deleteitem(args)
                else:
                    raise NotImplementedError

        return out

    def do_batchgetitem(self, args):
        out = {'Responses': {}}
        for table_name, requests in args['RequestItems'].iteritems():
            out['Responses'].setdefault(table_name, {'Items': []})
            for key in requests['Keys']:
                # We'll go through and get each item as long as our capcity can handle it.
                if not self.tables[table_name]['read_counter'].check():
                    out.setdefault('UnprocessedKeys', {})
                    out['UnprocessedKeys'].setdefault(table_name, {'Keys': []})
                    out['UnprocessedKeys'][table_name]['Keys'].append(key)
                else:
                    # TODO: This would probably be nice as a better multi-key query
                    item = self._get_item(table_name, key)
                    if item:
                        out['Responses'][table_name]['Items'].append(item)

        return out

    def do_scan(self, args):
        table_spec = self.tables[args['TableName']]
        sql_table = table_spec['sql_table']

        q = sql.select([sql_table])
        if 'Limit' in args:
            q = q.limit(args['Limit'])

        unsupported_keys = set(args.keys()) - set(['Limit', 'TableName', 'ScanFilter']) 
        if unsupported_keys:
            raise NotImplementedError(unsupported_keys)

        capacity_size = 0
        out = {'Items': []}
        for res in self.engine.execute(q):
            capacity_size += (len(res[sql_table.c.content]) / 1024) + 1
            item_data = json.loads(res[sql_table.c.content])
            item = utils.parse_item(item_data)
            if 'ScanFilter' in args:
                for attribute, filter_spec in args['ScanFilter'].iteritems():
                    if attribute not in item:
                        continue
                    for value_spec in filter_spec['AttributeValueList']:
                        value = utils.parse_value(value_spec)
                        if compare(filter_spec['ComparisonOperator'], item[attribute], value):
                            out['Items'].append(item_data)
                            break
            else:
                out['Items'].append(item_data)

        self.table_read_op_capacity[args['TableName']] += capacity_size
        self.tables[args['TableName']]['read_counter'].record(capacity_size)

        return out

    def do_query(self, args):
        table_spec = self.tables[args['TableName']]
        sql_table = table_spec['sql_table']
        key_spec = table_spec['key_spec']
        table_def = table_spec['table_def']

        unsupported_keys = set(args.keys()) - set(['Limit', 'TableName', 'HashKeyValue', 'ScanIndexForward', 'ConsistentRead', 'RangeKeyCondition', 'AttributesToGet', 'ExclusiveStartKey']) 
        if unsupported_keys:
            raise NotImplementedError(unsupported_keys)

        scan_forward = args.get('ScanIndexForward', True)

        expr = sql_table.c.hash_key == utils.parse_value(args['HashKeyValue'])

        if 'RangeKeyCondition' in args:
            if len(key_spec) < 2:
                raise NotImplementedError

            operator = args['RangeKeyCondition']['ComparisonOperator']
            if operator == 'BETWEEN':
                start = utils.parse_value(args['RangeKeyCondition']['AttributeValueList'][0])
                end = utils.parse_value(args['RangeKeyCondition']['AttributeValueList'][1])
                expr = sql.and_(expr, sql_table.c.range_key.between(start, end))

            else:
                range_value = utils.parse_value(args['RangeKeyCondition']['AttributeValueList'][0])
                if operator == 'GT':
                    expr = sql.and_(expr, sql_table.c.range_key > range_value)
                elif operator == 'LT':
                    expr = sql.and_(expr, sql_table.c.range_key < range_value)
                elif operator == 'GE':
                    expr = sql.and_(expr, sql_table.c.range_key >= range_value)
                elif operator == 'LE':
                    expr = sql.and_(expr, sql_table.c.range_key <= range_value)
                elif operator == 'EQ':
                    expr = sql.and_(expr, sql_table.c.range_key == range_value)
                elif operator == 'BEGINS_WITH':
                    expr = sql.and_(expr, sql_table.c.range_key.like('%s%%' % range_value))
                else:
                    raise NotImplementedError

        if 'ExclusiveStartKey' in args:
            range_key = utils.parse_value(args['ExclusiveStartKey']['RangeKeyElement'])
            if scan_forward:
                expr = sql.and_(expr, sql_table.c.range_key > range_key)
            else:
                expr = sql.and_(expr, self.table.c.range_key < range_key)

        q = sql.select([sql_table], expr)

        if 'Limit' in args:
            limit = min(DEFAULT_LIMIT, args['Limit'])
        else:
            limit = DEFAULT_LIMIT

        q = q.limit(limit + 1)

        if len(key_spec) > 1:
            if scan_forward:
                q = q.order_by(sql_table.c.range_key.asc())
            else:
                q = q.order_by(sql_table.c.range_key.desc())

        capacity_size = 0
        out = {'Items': [], 'Count': 0}
        for res in self.engine.execute(q):
            if out['Count'] == limit:
                break

            capacity_size += (len(res[sql_table.c.content]) / 1024) + 1

            item_data = json.loads(res[sql_table.c.content])
            item = utils.parse_item(item_data)
            if 'AttributesToGet' in args:
                out_item = dict((col_name, item_data[col_name]) for col_name in args['AttributesToGet'])
                out['Items'].append(out_item)
            else:
                out['Items'].append(item_data)

            out['Count'] += 1
            if len(key_spec) > 1:
                out['LastEvaluatedKey'] = utils.format_item({'HashKeyElement': res[sql_table.c.hash_key], 'RangeKeyElement': res[sql_table.c.range_key]})
        else:
            # If we didn't break out of our loop, that means we're on the last page of our result set and should have a key

            # The docs say the last evaulated key should be 'null', but I'm suspicious that isn't the case.
            if 'LastEvaluatedKey' in out:
                #out['LastEvaluatedKey'] = None
                del out['LastEvaluatedKey']
            

        self.table_read_op_capacity[args['TableName']] += capacity_size
        self.tables[args['TableName']]['read_counter'].record(capacity_size)

        out['Count'] = len(out['Items'])

        return out

    def make_request(self, command, body=None, callback=None):
        self.reset_op_capacity() 

        args = json.loads(body)
        # Record and check our provisioning limits
        if 'TableName' in args:
            if command in ('GetItem', 'BatchGetItem') and not self.tables[args['TableName']]['read_counter'].check():
                return callback({}, error=errors.ProvisionedThroughputError())

            if command in ('PutItem', 'BatchWriteItem', 'DeleteItem', 'UpdateItem') and not self.tables[args['TableName']]['write_counter'].check():
                return callback({}, error=errors.ProvisionedThroughputError())

        result = None
        try:
            result = getattr(self, "do_%s" % command.lower())(args)
        except CommandError, e:
            callback({}, error=e.args[0])
            return
        else:
            if result is None:
                result = {}

            # Record and check our provisioning limits
            if 'TableName' in args:
                if self.table_read_op_capacity:
                    for _, capacity in self.table_read_op_capacity.iteritems():
                        result['ConsumedCapacityUnits'] = float(capacity)

                elif self.table_write_op_capacity:
                    for _, capacity in self.table_write_op_capacity.iteritems():
                        result['ConsumedCapacityUnits'] = float(capacity)

            else:
                for table_name, capacity in self.table_read_op_capacity.iteritems():
                    result.setdefault('Responses', {}).setdefault(table_name, {})['ConsumedCapacityUnits'] = float(capacity)

                for table_name, capacity in self.table_write_op_capacity.iteritems():
                    result.setdefault('Responses', {}).setdefault(table_name, {})['ConsumedCapacityUnits'] = float(capacity)

            callback(result, error=None)


class SQLDB(db.BaseDB):
    def __init__(self, engine):
        super(SQLDB, self).__init__()
        self.client = SQLClient(engine)

    def register(self, table, create=False):
        super(SQLDB, self).register(table)
        if issubclass(table, db.Table):
            self.client.register_table(table)


if __name__  == '__main__':
    engine = sqlalchemy.create_engine('sqlite:///:memory:', echo=True)

    db = SQLDB(engine, 'RhettTest', ('user',))

    db['Britt'] = {'last_name': 'Deal', 'value': 9}
    db['Rhett'] = {'last_name': 'Rhett', 'value': 10}

    print db['Britt']

    #s = db.scan().limit(2) #.filter_gt('value', 10)
    #for r in s():
        #print r

    q = db.query('Britt').reverse().attributes(['user', 'value']).defer()
    res, error = q()
    for r in res:
        print r

    #print db[('Rhett', 'A')]




# -*- coding: utf-8 -*-
"""
This module contains the Solvent system which allows for higher level database operations
on a DynamoDB data store.

Usage is generally to create a instance of Solvent() class, configuring it with a set of operations,
and then running it.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""
import time
import logging
import copy
import types

from . import operation
from . import view
from . import errors
from . import db

log = logging.getLogger(__name__)

MAX_ATTEMPTS = 5

def classify(table_or_cls):
    if table_or_cls.__class__ == type:
        return table_or_cls
    else:
        return table_or_cls.__class__

class Solvent(object):
    """A solvent is a abstraction over Dynochemy database operations where
    operations can be combined together and executed with some intelligence.

    This includes:
        * automatic secondary index maintenance
        * throttling
        * memcache write-through caches and invalidation
    """
    def __init__(self):
        # We always start with an operation set, because everything can be reduced into it.
        self.operations = []

    def add_operation(self, op):
        self.operations.append(op)

    def __len__(self):
        return len(self.operations)

    def put(self, table, entity):
        table_cls = classify(table)
        op = operation.PutOperation(table_cls, entity)
        self.add_operation(op)
        return op

    def delete(self, table, key, noindex=False):
        table_cls = classify(table)
        op = operation.GetAndDeleteOperation(table_cls, key, noindex=noindex)
        self.add_operation(op)
        return op

    def update(self, table, key, put=None, add=None, delete=None):
        table_cls = classify(table)
        op = operation.UpdateOperation(table_cls, key, put=put, add=add, delete=delete)
        self.add_operation(op)
        return op

    def get(self, table, key):
        table_cls = classify(table)
        op = operation.GetOperation(table_cls, key)
        self.add_operation(op)
        return op

    def query(self, table_or_view, key):
        table_or_view_cls = classify(table_or_view)
        if issubclass(table_or_view_cls, db.Table):
            op = operation.QueryOperation(table_or_view_cls, key)
            self.add_operation(op)
        else:
            op = table_or_view_cls.query_op(key)
            self.add_operation(op)

        return op

    def scan(self, table):
        raise NotImplementedError

    def run(self, db):
        df = self.run_defer(db)
        result = df()
        return result

    def run_async(self, db, callback=None):
        def handle_result(cb):
            if callback is not None:
                result = cb.result
                callback(result)

        df = self.run_defer(db)
        df.add_callback(handle_result)

    def run_defer(self, db):
        run = SolventRun(db, self.operations)
        return run.run_defer()


class SolventRun(object):
    def __init__(self, db, ops):
        self.db = db

        self.operations = copy.copy(ops)

        self.op_results = operation.OperationResult(db)
        self.op_results.add_callback(self.handle_result)

        self.defer = operation.OperationResultDefer(self.op_results, db.ioloop)

        self.current_op_dfs = []

    def add_operation(self, op):
        self.operations.append(op)

    def run_defer(self):
        self.next_step()

        return self.defer

    def handle_result(self, op):
        try:
            result = self.op_results[op]
        except Exception:
            return

        if op.noindex:
            return

        for view in self.db.views_by_table(op.table):
            next_ops = view.operations_for_operation(op, result)
            if next_ops:
                log.debug("Found %d view operations for %r", len(next_ops), op)

            for new_op in next_ops:
                self.add_operation(new_op)

    def next_step(self):
        next_ops = []
        # Start our set of operations off with whatever our results object says we have to do.
        # These are usually complex operations with multiple stages.
        while self.op_results.next_ops:
            next_ops.append(self.op_results.next_ops.pop(0))

        # No ops left over from last time, grab the next part of the sequence.
        if not next_ops and self.operations:
            next_ops += self.operations
            del self.operations[:]

        if not next_ops:
            # All done, no more operations to complete
            self.defer.callback()
            return

        for op in OperationSet(next_ops).ops:
            op_df = op.run_defer(self.op_results)
            self.current_op_dfs.append(op_df)

        # We add our callbacks all at once, because in sync mode, we want to
        # ensure all our op_dfs have been created before we start to look for
        # finished ones.
        for op_df in self.current_op_dfs:
            # After each op, we also need to check for completion of all our
            # ops (based on the all_op_dfs list itself)
            op_df.add_callback(self.check_step_done)


    def check_step_done(self, cb):
        """Check if our solvent has completed a set of operations and handle errors"""
        if self.current_op_dfs and all(df.done for df in self.current_op_dfs):

            # We have a current list, and they are all done. So this guy is complete.
            del self.current_op_dfs[:]

            has_failed_ops = self.requeue_failed_ops()
            if has_failed_ops:
                delay_secs = 0.8 * self.op_results.error_attempts
                log.info("Op failures, trying again in %.1f seconds", delay_secs)

                db = self.op_results.db
                if db.ioloop:
                    db.ioloop.add_timeout(time.time() + delay_secs, self.next_step)
                else:
                    # No ioloop, do it inline
                    time.sleep(delay_secs)
                    self.next_step()
            else:
                self.next_step()

    def requeue_failed_ops(self):
        has_failures = False
        error_tables = set()
        op_error_count = 0

        # Check for errors and do some retries.
        for op in self.op_results:
            err = self.op_results.raw_error(op)

            # Certain types of errors we know we can try again just be re-executing the same operation.
            # Other errors, or successes, we'll just record and carry on.
            if isinstance(err, (errors.ProvisionedThroughputError, errors.UnprocessedItemError)):
                log.debug("Provisioning error for %r on table %r", op, op.table.name)

                # We'll only requeue failed operations a fixed number of times.
                if self.op_results.error_attempts < MAX_ATTEMPTS:
                    self.add_operation(op)
                    has_failures = True
                    op_error_count += 1
                    error_tables.add(op.table.name)

        if has_failures:
            self.op_results.error_attempts += 1
            log.warning("(%d) Retrying %d operations on tables %r due to provisioning limits", self.op_results.error_attempts, op_error_count, list(error_tables))
        else:
            self.op_results.error_attempts = 0

        return has_failures


class OperationSet(object):
    """Object for combining multiple operations together intelligently.

    Individual operations such as Put and Get and some Updates can be combined together into batches.
    """
    def __init__(self, operation_list=None):
        # Really we can combine everything into into two operations
        self.update_ops = []
        self.other_ops = []
        self.batch_write_op = operation.BatchWriteOperation()
        self.batch_read_op = operation.BatchReadOperation()

        if operation_list:
            for op in operation_list:
                self.add(op)

    def __len__(self):
        return len(self.update_ops) + len(self.other_ops) + len(self.batch_write_op) + len(self.batch_read_op)

    def __copy__(self):
        return OperationSet(list(self))

    def __iter__(self):
        for op in itertools.chain(self.update_ops, self.other_ops, self.batch_write_op, self.batch_read_op):
            yield op

    @property
    def ops(self):
        ops = []
        ops += self.update_ops
        ops += self.other_ops
        if len(self.batch_write_op) > 0:
            ops.append(self.batch_write_op)
        if len(self.batch_read_op) > 0:
            ops.append(self.batch_read_op)

        return ops

    def add(self, op):
        if isinstance(op, operation._WriteBatchableMixin):
            self.batch_write_op.add(op)
        elif isinstance(op, operation._ReadBatchableMixin):
            self.batch_read_op.add(op)
        elif isinstance(op, operation.UpdateOperation):
            self.add_update(op)
        elif isinstance(op, operation.QueryOperation):
            self.other_ops.append(op)
        else:
            raise ValueError("Don't know how to add op %r" % op)

    def add_update(self, update_op):
        # We could keep just a list of all updates and then execute them
        # sequentially, but updates might be easily combined if they are for the same
        # table/key pair

        combined_update_op = False
        new_update_ops = []
        for op in self.update_ops:
            if not combined_update_op and op.table == update_op.table and op.key == update_op.key:
                # We can combine these updates
                new_update_ops.append(op.combine_updates(update_op))
                combined_update_op = True
            else:
                new_update_ops.append(op)

        if not combined_update_op:
            new_update_ops.append(update_op)

        self.update_ops = new_update_ops
        return


if __name__ == '__main__':

    class TestTable(Table):
        name = 'test_table'
        hash_key = 'key'

    db.register(TestTable)

    s = TestSolvent()

    put_op_1 = s.put(TestTable, {'key': 'hello', 'value': 10})
    put_op_2 = s.put(TestTable, {'key': 'world', 'value': 25.2})

    result = s.run(db)

    print result[put_op_1]
    print result[put_op_2]



# -*- coding: utf-8 -*-

"""
dynochemy.utils
~~~~~~~~

This module provides utility functions that are used within Dynochemy
especially for things like formatting datastructures for Dynamo.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.

"""
import time
import logging
import itertools
import types
import contextlib

from tornado.simple_httpclient import _HTTPConnection
from tornado.httpclient import HTTPResponse
from tornado.ioloop import IOLoop

log = logging.getLogger(__name__)

def stringify(val):
    if isinstance(val, bool):
        return str(int(val))
    return str(val)

def format_value(value):
    if isinstance(value, basestring):
        if not value:
            raise ValueError("empty string")
        return {'S': value}
    elif isinstance(value, (int, long, float, bool)):
        return {'N': stringify(value)}
    elif isinstance(value, (list, tuple, set)):
        all_values = [format_value(v) for v in set(value)]
        spec = None 
        values = []
        for value_spec in all_values:
            if value_spec.keys()[0] == 'S':
                spec = 'SS'
            elif value_spec.keys()[0] == 'N':
                spec = 'NS'
            values.append(value_spec.values()[0])
        return {spec: values}
    else:
        raise ValueError(value)

def _parse_value_spec(type_spec, value):
    if type_spec == 'S':
        return value
    if type_spec == 'N' and '.' in value:
        return float(value)
    elif type_spec == 'N':
        return int(value)
    else:
        raise ValueError(type_spec)

def format_key(key_spec, key_value):
    assert isinstance(key_value, (list, tuple))
    if len(key_spec) == 1:
        return {key_spec[0]: format_value(key_value[0])}
    else:
        out_key = {}
        for key, value in zip(key_spec, key_value):
            out_key[key] = format_value(value)
        return out_key

def parse_key(key):
    parsed_value = parse_item(key)

    if 'RangeKeyElement' in parsed_value:
        return (parsed_value['HashKeyElement'], parsed_value['RangeKeyElement'])
    else:
        return (parsed_value['HashKeyElement'],)

EMPTY_VALUES = ['', None, [], set(), tuple()]
def format_item(item):
    try:
        out = {}
        for k, v in item.iteritems():
            if v in EMPTY_VALUES:
                continue
            out[k] = format_value(v)
    except (TypeError, ValueError), e:
        log.error("Failed to format value for %s", k)
        raise
    return out

def parse_value(value_spec):
    key, value = value_spec.items()[0]
    if len(key) == 1:
        return _parse_value_spec(key, value)
    elif key == 'SS':
        # Note that we could use 'sets' for these, but they are kinda awkward
        # to deal with (and don't natively json encode)
        return value
    elif key == 'NS':
        return [_parse_value_spec('N', v) for v in value]
    else:
        raise ValueError(key)

def parse_item(item):
    return dict((k, parse_value(v)) for k, v in item.iteritems())

def segment(iterable, size):
    it = iter(iterable)
    while True:
        group = tuple(itertools.islice(it, None, size))
        if not group:
            break
        yield group


class ResourceCounter(object):
    """Utility for tracking limits in resource usage over time.

    We use this to track, how much capacity has been used for reading or writing to DynamoDB.

    General usage is to:
        * Create TimeCounter with a size in seconds (how much time we want to track)
        * Set limits over past n seconds with set_limit()
        * Record values as they occur. These will be slotted with 1 second accuracy.
        * Call check() to see if limits have been exceeded. 
        * Call avg() to see what the moving average is over the specified interval.
    """
    def __init__(self, size=60):
        self.size = size
        self.last_second = None
        self.values = []
        self.limits = []

    def avg(self, seconds=1):
        """Returns the moving average over the specified interval"""
        if self.last_second is None:
            return 0.0

        age = int(time.time()) - self.last_second
        interval_left = seconds - age

        if seconds - age > 0:
            value = sum(self.values[0:interval_left]) / float(seconds)
            log.debug("Average over %d: %.1f", seconds, value)
            return value

        return 0.0

    def record(self, value):
        current_time = int(time.time())
        if self.last_second is None:
            self.values.append(0)
        elif self.last_second < current_time:
            while self.last_second < current_time:
                self.values.insert(0, 0)
                self.last_second += 1

        self.values[0] += value
        self.last_second = current_time
        del self.values[self.size:]

    def check(self):
        for seconds, value in self.limits:
            if self.avg(seconds) > value:
                return False

        return True

    def set_limit(self, seconds, value):
        self.limits.append((seconds, value))


def predict_capacity_usage(request_type, args, result=None):
    if request_type == 'GetItem':
        return 1.0
    elif request_type == 'BatchGetItem':
        out = {}
        for table_name, requests in args['RequestItems'].iteritems():
            out[table_name] = 1.0 * len(requests['Keys'])
        return out
    elif request_type == 'PutItem':
        return 1.0
    elif request_type == 'UpdateItem':
        return 1.0
    elif request_type == 'DeleteItem':
        return 1.0
    elif request_type == 'BatchWriteItem':
        out = {}
        for table_name, requests in args['RequestItems'].iteritems():
            out[table_name] = 1.0 * len(requests)
        return out
    elif request_type == 'Query':
        if result:
            return 1.0 * len(result['Items'])
        else:
            return 25.0
    elif request_type == 'Scan':
        if result:
            return 1.0 * len(result['Items'])
        else:
            return 25.0
    else:
        raise ValueError(request_type)


def patch_io_loop():
    """Create an IOLoop sub-class

    We have out own io loop subclass that handles unhandled exceptions by
    re-raising them and quitting.  This is useful for io loops used in contexts
    that are not multi-user... for example dynochemy uses its own ioloops for
    parallelizing io operations, in which case an exception would be better off
    just raised out of ioloop.start()

    This is a function that generates a class because newwer version of tornado
    has some dynamic IOLoop sub-class selection mechanism based on platform. So
    we can just statically subclass 'IOLoop'.
    """
    if hasattr(IOLoop, 'configurable_default'):
        base_cls = IOLoop.configurable_default()
    else:
        base_cls = IOLoop

    class StrictExceptionIOLoop(base_cls):
        def handle_callback_exception(self, callback):
            raise

    return StrictExceptionIOLoop

def patch_http_client():
    """Patch a simple_httpclient instance to raise exceptions generated in callbacks.

    See https://github.com/facebook/tornado/pull/652
    """
    @contextlib.contextmanager
    def better_cleanup(self):
        try:
            yield
        except Exception, e:
            if self.final_callback:
                self._run_callback(HTTPResponse(self.request, 599, error=e,
                                   request_time=time.time() - self.start_time,
                                ))

                if hasattr(self, "stream"):
                    self.stream.close()
            else:
                # If our callback has already been called, we are probably
                # catching an exception that is not caused by us but rather
                # some child of our callback. Rather than drop it on the floor,
                # pass it along.
                raise

    _HTTPConnection.cleanup = types.UnboundMethodType(better_cleanup, None, _HTTPConnection)




# -*- coding: utf-8 -*-
"""
This module contains the primary objects that powers Dynochemy.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""
import copy
import json
import time
import logging
import pprint
import collections

from asyncdynamo import asyncdynamo

from .errors import Error, SyncUnallowedError, DuplicateBatchItemError, UnprocessedItemError, ExceededBatchRequestsError, ItemNotFoundError, parse_error
from . import utils
from .defer import Defer
from . import view
from . import constants
from . import utils


log = logging.getLogger(__name__)


class BaseDB(object):
    def __init__(self):
        self.allow_sync = True
        self.ioloop = None
        self.tables = {}
        self.views = {}
        self._tables_by_db_name = {}
        self._views_by_table = collections.defaultdict(list)

    def register(self, cls, create=False):
        if issubclass(cls, Table):
            self.tables[cls.__name__] = instance = cls(self)
            self._tables_by_db_name[cls.name] = instance
        elif issubclass(cls, view.View):
            self.views[cls.__name__] = instance = cls(self)
            self._views_by_table[cls.table.__name__].append(instance)

    def __getattr__(self, name):
        try:
            return self.tables[name]
        except KeyError:
            try:
                return self.views[name]
            except KeyError:
                raise AttributeError(name)

    def table_by_name(self, name):
        return self._tables_by_db_name[name]

    def views_by_table(self, table):
        return self._views_by_table[table.__name__]

    def batch_write(self):
        return WriteBatch(self)

    def batch_read(self):
        return ReadBatch(self)


class DB(BaseDB):
    def __init__(self, access_key, access_secret, ioloop=None):
        super(DB, self).__init__()

        self.allow_sync = bool(ioloop is None)
        self.ioloop = ioloop or utils.patch_io_loop()()

        # This will monkey patch the _HTTPConnection class in tornado
        # Eventually when the real patch is released, we should do some sort of version check.
        utils.patch_http_client()

        self.client = asyncdynamo.AsyncDynamoDB(access_key, access_secret, ioloop=self.ioloop)


class Table(object):
    name = None
    hash_key = None
    range_key = None

    read_capacity = None
    write_capacity = None

    def __init__(self, db):
        self.db = db
        self.read_counter = utils.ResourceCounter()
        self.write_counter = utils.ResourceCounter()

    @property
    def key_spec(self):
        if self.range_key:
            return (self.hash_key, self.range_key)
        else:
            return (self.hash_key,)

    def _record_read_capacity(self, value):
        log.debug("%.1f read capacity units consumed", value)
        self.read_counter.record(value)

    def _record_write_capacity(self, value):
        log.debug("%.1f write capacity units consumed", value)
        self.write_counter.record(value)

    def _get(self, key, attributes=None, consistent=True):
        data = {
                'TableName': self.name,
               }

        if self.has_range:
            data['Key'] = utils.format_key(('HashKeyElement', 'RangeKeyElement'), key)
        else:
            data['Key'] = utils.format_key(('HashKeyElement',), (key,))

        if attributes:
            data['AttributesToGet'] = attributes

        data['ConsistentRead'] = consistent

        defer = Defer(self.db.ioloop)

        def handle_result(data, error=None):
            if error is not None:
                defer.exception(ex=parse_error(error))
                return

            read_capacity = {self.name: 0.0}
            if 'ConsumedCapacityUnits' in data:
                read_capacity[self.name] = float(data['ConsumedCapacityUnits'])
                self._record_read_capacity(read_capacity[self.name])

            if 'Item' in data:
                defer.callback(utils.parse_item(data['Item']), read_capacity=read_capacity)
            else:
                defer.callback(None, read_capacity=read_capacity)

        self.db.client.make_request('GetItem', body=json.dumps(data), callback=handle_result)
        return defer

    def get_async(self, key, callback, attributes=None, consistent=True):
        df = self._get(key, attributes=attributes, consistent=consistent)

        def handle_result(get_df):
            callback(get_df.result)

        df.add_callback(handle_result)

    def get_defer(self, key, attributes=None, consistent=True):
        return self._get(key, attributes=attributes, consistent=consistent)

    def __getitem__(self, key):
        if not self.db.allow_sync:
            raise SyncUnallowedError()

        d = self._get(key)

        args, kwargs = d()

        if not args or args[0] is None:
            raise KeyError(key)

        return args[0]

    get = __getitem__

    def _put(self, value):
        data = {
                'TableName': self.name,
               }

        item = utils.format_item(value)
        data['Item'] = item

        defer = Defer(self.db.ioloop)

        def handle_result(data, error=None):
            if error is not None:
                defer.exception(ex=parse_error(error))
            else:
                write_capacity = {self.name: 0.0}
                if 'ConsumedCapacityUnits' in data:
                    write_capacity[self.name] = float(data['ConsumedCapacityUnits'])
                    self._record_write_capacity(write_capacity[self.name])

                defer.callback(None, write_capacity=write_capacity)

        self.db.client.make_request('PutItem', body=json.dumps(data), callback=handle_result)
        return defer

    def put_async(self, value, callback):
        df = self._put(value)

        def handle_result(put_df):
            callback(put_df.result)

        df.add_callback(handle_result)

    def put_defer(self, value):
        return self._put(value)

    def __setitem__(self, key, value):
        item = copy.copy(value)
        if len(self.key_spec) > 1:
            for key, key_value in zip(self.key_spec, key):
                item[key] = key_value
        else:
            item[self.key_spec[0]] = key

        self.put(item)

    def put(self, value, timeout=None):
        if not self.db.allow_sync:
            raise SyncUnallowedError()

        d = self._put(value)

        args, kwargs = d(timeout=timeout)
        return args and args[0]

    def _delete(self, key):
        data = {
                'TableName': self.name,
               }

        if self.has_range:
            data['Key'] = utils.format_key(('HashKeyElement', 'RangeKeyElement'), key)
        else:
            data['Key'] = utils.format_key(('HashKeyElement',), (key,))

        data['ReturnValues'] = "ALL_OLD"

        defer = Defer(self.db.ioloop)

        def handle_result(data, error=None):
            if error is not None:
                defer.exception(ex=parse_error(error))
                return

            write_capacity = {self.name: 0.0}
            if 'ConsumedCapacityUnits' in data:
                write_capacity[self.name] = float(data['ConsumedCapacityUnits'])
                self._record_write_capacity(write_capacity[self.name])

            if 'Attributes' in data:
                defer.callback(utils.parse_item(data['Attributes']), write_capacity=write_capacity)
            else:
                defer.callback(None, write_capacity=write_capacity)

        self.db.client.make_request('DeleteItem', body=json.dumps(data), callback=handle_result)
        return defer

    def delete_async(self, key, callback):
        df = self._delete(key)

        def handle_result(del_df):
            callback(del_df.result)

        df.add_callback(handle_result)

    def delete_defer(self, key):
        return self._delete(key)

    def __delitem__(self, key):
        if not self.db.allow_sync:
            raise SyncUnallowedError()

        d = self._delete(key)

        args, kwargs =  d()

        if not args or args[0] is None:
            raise KeyError(key)

        return args[0]

    delete = __delitem__

    def _update(self, key, add=None, put=None, delete=None, return_value=None):
        data = {
                'TableName': self.name,
                'ReturnValues': return_value or 'ALL_NEW',
               }

        if self.has_range:
            data['Key'] = utils.format_key(('HashKeyElement', 'RangeKeyElement'), key)
        else:
            data['Key'] = utils.format_key(('HashKeyElement',), [key])

        data['AttributeUpdates'] = {}

        if add:
            for attribute, value in add.iteritems():
                if value not in ('', None):
                    update = {attribute: {'Value': utils.format_value(value), 'Action': 'ADD'}}
                    data['AttributeUpdates'].update(update)
        if put:
            for attribute, value in put.iteritems():
                if attribute in self.key_spec:
                    raise ValueError("can't put key attributes")
                if value not in ('', None):
                    update = {attribute: {'Value': utils.format_value(value), 'Action': 'PUT'}}
                    data['AttributeUpdates'].update(update)

        if delete:
            for attribute, value in delete.iteritems():
                update = {attribute: {'Action': 'DELETE'}}
                if value is not None:
                    update[attribute]['Value'] = utils.format_value(value)
                data['AttributeUpdates'].update(update)

        defer = Defer(ioloop=self.db.ioloop)

        def handle_result(result, error=None):
            if error is not None:
                defer.exception(ex=parse_error(error))
                return

            write_capacity = 0.0
            if 'ConsumedCapacityUnits' in result:
                write_capacity = float(result['ConsumedCapacityUnits'])
                self._record_write_capacity(write_capacity)

            if 'Attributes' in result:
                defer.callback(utils.parse_item(result['Attributes']), write_capacity={self.name: write_capacity})
            else:
                defer.callback(None, write_capacity={self.name: write_capacity})

        self.db.client.make_request('UpdateItem', body=json.dumps(data), callback=handle_result)
        return defer

    def update(self, key, add=None, put=None, delete=None, return_value=None, timeout=None):
        if not self.db.allow_sync:
            raise SyncUnallowedError()

        d = self._update(key, add=add, put=put, delete=delete, return_value=return_value)

        args, kwargs = d(timeout=timeout)

        return args[0]

    def update_defer(self, key, add=None, put=None, delete=None, return_value=None, timeout=None):
        return self._update(key, add=add, put=put, delete=delete, return_value=return_value)

    def update_async(self, key, callback, add=None, put=None, delete=None):
        df = self._update(key, add=add, put=put, delete=delete)

        def handle_result(update_df):
            callback(update_df.result)

        df.add_callback(handle_result)

    def scan(self):
        return Scan(self)

    def scan_next(self, result):
        if not result.has_next:
            raise ValueError("Result has no more")

        scan = copy.copy(result.scan)
        scan.args['ExclusiveStartKey'] = result.result_data['LastEvaluatedKey']
        return scan

    def query(self, hash_key):
        return Query(self, hash_key)

    def query_next(self, result):
        if not result.has_next:
            return None

        query = copy.copy(result.query)
        query.args['ExclusiveStartKey'] = result.result_data['LastEvaluatedKey']
        return query

    @property
    def has_range(self):
        return bool(len(self.key_spec) == 2)

    def _item_key(self, item):
        """Return a tuple of identifying information for the item.

        This is based on the keys."""
        key = [utils.parse_value(item[k]) for k in self.key_spec]
        return tuple(key)

    def _key_key(self, value):
        """Return a tuple of identifying information for the parsed key"""

        key = [utils.parse_value(value[name]) for name, _ in zip(('HashKeyElement', 'RangeKeyElement'), self.key_spec)]
        return tuple(key)


class Batch(object):
    """Object for doing a batch operation.

    Usage should be something like:

        b = db.batch_write()
        b.put(item1)
        b.put(item2)

        b()

        if b.errors:
            raise Exception

    Each operation returns a 'defer' object that can later retrieve results.
    Note that a batch is in no way atomic. Even the underlying batch operations provided by DynamoDB are not
    really atomic.

    Typically, a request will either error or not. However, some elements may be unprocessed due to exceeding capacity.
    In this case, the defer object for the operation will not be complete (`not d.done`).
    """

    def __init__(self, db):
        self.db = db

        # We use several data structures to manage our requests
        # Requests are identified by a key, the format varying based on type of request. Generally it's a tuple
        # and some uniquly identifying information.

        # Our requests that need to be processed
        self._requests = []

        # Map of request key to the actual request data
        # TODO: needed?
        self._request_data = {}

        # Map of request key to the defer we returned for it.
        self._request_defer = {}

        # This defer tracks the over all status of this batch.
        self._defer = Defer(ioloop=self.db.ioloop)

        self.errors = []

    def __getattr__(self, name):
        if hasattr(self.db, name):
            tbl = getattr(self.db, name)
            return BatchTable(self, tbl)
        else:
            raise AttributeError

    def __call__(self, timeout=None):
        self._run_batch()

        args, kwargs = self._defer(timeout=timeout)
        return args[0]

    def async(self, callback=None):
        if callback:
            def defer_callback(df):
                callback(df.result)
            self._defer.add_callback(defer_callback)

        self._run_batch()

    def defer(self):
        self._run_batch()
        return self._defer

    def _make_request(requests):
        raise NotImplementedError


class BatchTable(object):
    def __init__(self, batch, table):
        self.table = table
        self.batch = batch

    def put(self, value):
        return self.batch.put(self.table, value)

    def delete(self, key):
        return self.batch.delete(self.table, key)

    def get(self, key):
        return self.batch.get(self.table, key)


class WriteBatch(Batch):
    def put(self, table, value):
        df = Defer(ioloop=self._defer.ioloop)
        args = {'PutRequest': {"Item": utils.format_item(value)}}

        req_key = (table.name, "PutRequest", table._item_key(args['PutRequest']['Item']))
        if req_key in self._request_defer:
            raise DuplicateBatchItemError(value)

        log.debug("Building request %r", req_key)

        if len(self._requests) >= constants.MAX_BATCH_WRITE_ITEMS:
            raise ExceededBatchRequestsError("Too many requests")

        self._request_defer[req_key] = df
        self._request_data[req_key] = (table.name, args)
        self._requests.append(req_key)

        return df

    def delete(self, table, key):
        df = Defer(ioloop=self._defer.ioloop)
        req_args = {}
        if table.has_range:
            req_args['Key'] = utils.format_key(('HashKeyElement', 'RangeKeyElement'), key)
        else:
            req_args['Key'] = utils.format_key(('HashKeyElement',), (key,))

        args = {'DeleteRequest': req_args}

        req_key = (table.name, "DeleteRequest", key)

        log.debug("Building request %r", req_key)

        if len(self._requests) >= constants.MAX_BATCH_WRITE_ITEMS:
            raise ExceededBatchRequestsError("Too many requests")

        self._request_defer[req_key] = df
        self._request_data[req_key] = (table.name, args)
        self._requests.append(req_key)

        return df

    def _run_batch(self):
        log.debug("Creating BatchWrite request for %d items", len(self._requests))

        args = {"RequestItems": {}}

        for key in self._requests:
            table_name, req = self._request_data[key]
            args['RequestItems'].setdefault(table_name, []).append(req)

        def handle_result(data, error=None):
            if error is not None:
                real_error = parse_error(error)
                log.warning("Received error for batch: %r", real_error)

                for key in self._requests:
                    self._request_defer[key].exception(ex=real_error)

                self.errors.append(real_error)
                self._defer.exception(ex=real_error)
            else:
                log.debug("Received successful result from batch: %r", data)

                write_capacity = collections.defaultdict(float)
                if 'Responses' in data:
                    for table_name, response_data in data['Responses'].iteritems():
                        write_capacity[table_name] += float(response_data['ConsumedCapacityUnits'])
                        self.db.table_by_name(table_name)._record_write_capacity(float(response_data['ConsumedCapacityUnits']))

                unprocessed_items = set()
                if data.get('UnprocessedItems'):
                    for table_name, items in data['UnprocessedItems'].iteritems():
                        table = self.db.table_by_name(table_name)
                        for item in items:
                            (req_type, req), = item.items()

                            # Responses may be items or keys depending on the type of request
                            if 'Item' in req:
                                key = table._item_key(req['Item'])
                            else:
                                if table.has_range:
                                    key = table._key_key(req['Key'])
                                else:
                                    key = table._key_key(req['Key'])[0]

                            request = (table_name, req_type, key)

                            if request not in self._request_data:
                                log.warning("%r not found in %r", request, self._request_data.keys())
                                continue

                            self._request_defer[request].exception(ex=UnprocessedItemError())
                            unprocessed_items.add(request)

                if unprocessed_items:
                    log.warning("Found %d items unprocessed in BatchWrite", len(unprocessed_items))

                for key in self._requests:
                    if key not in unprocessed_items:
                        self._request_defer[key].callback(data)

                self._defer.callback(data, write_capacity=write_capacity)

        self.db.client.make_request('BatchWriteItem', body=json.dumps(args), callback=handle_result)


class ReadBatch(Batch):
    def get(self, table, key):
        df = Defer(ioloop=self._defer.ioloop)
        if table.has_range:
            req_key = utils.format_key(('HashKeyElement', 'RangeKeyElement'), key)
        else:
            req_key = utils.format_key(('HashKeyElement',), (key,))

        log.debug("Building request %r", req_key)

        if table.has_range:
            request = (table.name, key)
        else:
            request = (table.name, (key,))

        if len(self._requests) >= constants.MAX_BATCH_READ_ITEMS:
            raise ExceededBatchRequestsError("Too many requests")

        self._request_defer[request] = df
        self._request_data[request] = (table.name, req_key)
        self._requests.append(request)

        return df

    def _run_batch(self):
        log.debug("Creating ReadBatch request for %d items", len(self._requests))

        args = {"RequestItems": {}}

        for request in self._requests:
            table_name, key = self._request_data[request]
            args['RequestItems'].setdefault(table_name, {'Keys': []})['Keys'].append(key)

        def handle_result(data, error=None):
            if error is not None:
                real_error = parse_error(error)
                log.error("Received error for batch: %r", real_error)

                for request in self._requests:
                    self._request_defer[request].exception(ex=real_error)

                self._defer.exception(ex=real_error)
            else:
                log.debug("Received successful result from batch: %r", data)

                unprocessed_keys = set()
                if data.get('UnprocessedKeys'):
                    for table_name, unprocessed_items in data['UnprocessedKeys'].iteritems():
                        table = self.db.table_by_name(table_name)

                        for req_key in unprocessed_items['Keys']:
                            key = table._key_key(req_key)
                            request = (table_name, key)

                            if request not in self._request_data:
                                log.warning("%r not found in %r", request, self._request_data.keys())
                                continue

                            assert request in self._request_data
                            self._request_defer[request].exception(ex=UnprocessedItemError())
                            unprocessed_keys.add(request)

                if unprocessed_keys:
                    log.warning("Found %d keys unprocessed in BatchRead", len(unprocessed_keys))

                read_capacity = collections.defaultdict(float)
                for table_name, result in data['Responses'].iteritems():

                    if 'ConsumedCapacityUnits' in result:
                        read_capacity[table_name] += float(result['ConsumedCapacityUnits'])
                        self.db.table_by_name(table_name)._record_read_capacity(float(result['ConsumedCapacityUnits']))

                    for item in result['Items']:
                        table = self.db.table_by_name(table_name)
                        assert item
                        entity = utils.parse_item(item)
                        key = table._item_key(item)
                        request = (table_name, key)
                        self._request_defer[request].callback(entity)

                # Look for any items we didn't get results for.
                for req, df in self._request_defer.iteritems():
                    if not df.done:
                        self._request_defer[req].exception(ex=ItemNotFoundError())

                self._defer.callback(data, read_capacity=read_capacity)

        self.db.client.make_request('BatchGetItem', body=json.dumps(args), callback=handle_result)


class Scan(object):
    def __init__(self, table):
        self.table = table
        self.args = {
                     'TableName': self.table.name, 
                    }

    def limit(self, limit):
        scan = copy.copy(self)
        scan.args['Limit'] = limit
        return scan

    def _filter(self, name, value, compare):
        scan = copy.copy(self)
        scan.args.setdefault('ScanFilter', {})
        scan.args['ScanFilter'].setdefault(name, {})
        scan.args['ScanFilter'][name]['ComparisonOperator'] = compare
        scan.args['ScanFilter'][name].setdefault('AttributeValueList', [])
        scan.args['ScanFilter'][name]['AttributeValueList'].append(utils.format_value(value))
        return scan

    def filter_eq(self, name, value):
        return self._filter(name, value, "EQ")

    def filter_lt(self, name, value):
        return self._filter(name, value, "LT")

    def filter_gt(self, name, value):
        return self._filter(name, value, "GT")

    def _scan(self):
        defer = Defer(self.table.db.ioloop)

        def handle_result(data, error=None):
            if error is not None:
                defer.exception(ex=parse_error(error))
                return

            read_capacity = 0.0
            if 'ConsumedCapacityUnits' in data:
                read_capacity = float(data['ConsumedCapacityUnits'])
                self.table._record_read_capacity(read_capacity)

            defer.callback(ScanResults(self, data), read_capacity=read_capacity)

        self.table.db.client.make_request('Scan', body=json.dumps(self.args), callback=handle_result)
        return defer

    def __call__(self, timeout=None):
        if not self.table.db.allow_sync:
            raise SyncUnallowedError()

        d = self._scan()
        args, kwargs = d(timeout=timeout)
        return args[0]

    def defer(self):
        return self._scan()

    def async(self, callback=None):
        df = self._scan()

        def handle_result(scan_df):
            callback(scan_df.result)

        df.add_callback(handle_result)


class Query(object):
    def __init__(self, table, hash_key):
        self.table = table 
        self.args = {
                     'TableName': self.table.name, 
                     'HashKeyValue': utils.format_value(hash_key),
                     'ConsistentRead': True,
                    }

    def attributes(self, *args):
        query = copy.copy(self)
        query.args['AttributesToGet'] = args[0]
        return query
        
    def range(self, start=None, end=None):
        """Includes a condition where by range keys start at 'start' (inclusive) and are less than 'end' (exclusive)
        
        """
        if not self.table.range_key:
            raise ValueError("Must have range")

        query = copy.copy(self)
        query.args['RangeKeyCondition'] = condition = {'AttributeValueList': [], 'ComparisonOperator': None}
        if None not in (start, end):
            # We have a complete start to end condition
            condition['AttributeValueList'].append(utils.format_value(start))
            condition['AttributeValueList'].append(utils.format_value(end))
            condition['ComparisonOperator'] = "BETWEEN"

        elif start is not None:
            condition['AttributeValueList'].append(utils.format_value(start))
            condition['ComparisonOperator'] = "GE"

        elif end is not None:
            condition['AttributeValueList'].append(utils.format_value(end))
            condition['ComparisonOperator'] = "LT"

        return query

    def reverse(self, reverse=True):
        query = copy.copy(self)
        query.args['ScanIndexForward'] = not reverse
        return query
        
    def limit(self, limit):
        query = copy.copy(self)
        query.args['Limit'] = limit
        return query

    def last_seen(self, range_id):
        query = copy.copy(self)
        query.args['ExclusiveStartKey'] = {'HashKeyElement': self.args['HashKeyValue'], 'RangeKeyElement': utils.format_value(range_id)}
        return query

    def _query(self):
        defer = Defer(self.table.db.ioloop)

        def handle_result(result_data, error=None):
            results = None
            if error is None:
                if 'ConsumedCapacityUnits' in result_data:
                    self.table._record_read_capacity(float(result_data['ConsumedCapacityUnits']))

                defer.callback(QueryResults(self, result_data))
            else:
                defer.exception(ex=parse_error(error))

        self.table.db.client.make_request('Query', body=json.dumps(self.args), callback=handle_result)
        return defer

    def __call__(self, timeout=None):
        if not self.table.db.allow_sync:
            raise SyncUnallowedError()

        d = self._query()
        args, kwargs =  d(timeout=timeout)
        return args[0]

    def defer(self):
        return self._query()

    def async(self, callback=None):
        self._query(callback=callback)

        def handle_result(q_df):
            callback(q_df.result)

        df.add_callback(handle_result)


class Results(object):
    def __init__(self, result_data):
        self.result_data = result_data

    def __iter__(self):
        return (utils.parse_item(item) for item in self.result_data['Items'])

    def __len__(self):
        return self.result_data['Count']

    def __getitem__(self, key):
        return utils.parse_item(self.result_data['Items'][key])

    @property
    def has_next(self):
        return bool('LastEvaluatedKey' in self.result_data)


class QueryResults(Results):
    def __init__(self, query, result_data):
        self.query = query
        self.result_data = result_data

    def combine(self, other_results):
        return CombinedQueryResults(self, other_results)


class CombinedQueryResults(QueryResults):
    """Special version of QueryResults that puts multiple sets of QueryResults together.

    This is usually constructed by using the combine() method on a QueryResult.
    """
    def __init__(self, query_results, other_query_results):
        self.query = query_results.query
        self.result_data = {}

        self.result_data['Items'] = query_results.result_data['Items'] + other_query_results.result_data['Items']
        self.result_data['Count'] = query_results.result_data['Count'] + other_query_results.result_data['Count']


class SecondaryQueryResults(object):
    def __init__(self, query_results, table, op_results, key):
        self.query_results = query_results
        self.table = table
        self.op_results = op_results
        self.key_func = key

    def combine(self, other_results):
        self.query_results = self.query_results.combine(other_results)
        return self

    def __iter__(self):
        for item in self.query_results:
            key = self.key_func(item)
            op = operation.GetOperation(self.table, key)
            full_item, err = self.op_results[op]
            if err:
                raise err

            yield full_item

    def __len__(self):
        return len(self.query_results)


class ScanResults(Results):
    def __init__(self, scan, result_data):
        self.scan = scan
        self.result_data = result_data

def run_all(runnable):
    while True:
        result = runnable()
        for res in result:
            yield res

        if result.has_next:
            if isinstance(runnable, Scan):
                runnable = runnable.table.scan_next(result)
            else:
                runnable = runnable.table.query_next(result)
        else:
            break




# -*- coding: utf-8 -*-

"""
Dynochemy
~~~~~~~~

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.

"""

__title__ = 'dynochemy'
__version__ = '0.1.2'
__description__ = 'Clever pythonic and async interface to Amazon DynamoDB '
__url__ = 'https://github.com/rhettg/dynochemy'
__build__ = 0
__author__ = 'Rhett Garber'
__author_email__ = 'rhettg@gmail.com'
__license__ = 'ISC'
__copyright__ = 'Copyright 2012 Rhett Garber'


from . import utils
from .defer import Defer
from .db import DB
from .db import Table
from .db import run_all
from .view import View
from .solvent import Solvent
from .sql import SQLDB
from .oxdb import OxDB
from .oxdb import OxSQLDB
from .errors import *
from .operation import *



# -*- coding: utf-8 -*-
"""
This module contains objects for handling creating views in Dynochemy.

A view manages a table, populated from operations in another table. You can
think of this as indexing, but rather than just indexing entities, you can
create any kind of table you want. For example, you can have counts of entities
with certain values.

The only real requirement is that a view get registered for updates to a
specific table. It can then return additional operations that should be
included in the same batch of updates.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""
import copy

from . import operation


class View(object):
    # The table used as input for this view. The view will be informed by
    # updates to this table.
    table = None

    # What table are we writing to.
    view_table = None

    def __init__(self, db):
        # We allow instances of this to exist to provide some amount of parallelism with tables.
        # But i'm not yet quite sure if we really need to. All the important methods are classmethods
        self.db = db

    @classmethod
    def operations_for_operation(cls, op, result):
        if isinstance(op, operation.PutOperation):
            return cls.add(op, result)
        elif isinstance(op, operation.GetAndDeleteOperation):
            return cls.remove(op, result)
        elif isinstance(op, operation.UpdateOperation):
            return cls.update(op, result)
        elif isinstance(op, operation.DeleteOperation):
            # In a solvent, first we do a GetAndDeleteOperation, so this should have already been handled.
            pass
        elif isinstance(op, (operation.GetOperation, operation.QueryOperation)):
            pass
        else:
            raise NotImplementedError(op)

        return []

    @classmethod
    def add(cls, op, result):
        return []

    @classmethod
    def remove(cls, op, result):
        return []

    @classmethod
    def update(cls, op, result):
        return []





# -*- coding: utf-8 -*-
"""
This module contains tools for doing defered async operations.
Some of this is inspired by Twisted, but they are pretty different.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""
import functools
import time
import sys

from .errors import Error

class TimeoutError(Error): pass

class Defer(object):
    def __init__(self, ioloop=None):
        self.done = False
        self.ioloop = ioloop
        self.do_stop = False
        self.args = None
        self.kwargs = None
        self._timeout_req = None
        self._callbacks = []
        self._exception = None

    def add_callback(self, cb):
        self._callbacks.append(cb)
        if self.done:
            cb(self)

    def _do_done(self):
        if self.done:
            raise ValueError("Already called")

        self.done = True

        if self._timeout_req:
            self.ioloop.remove_timeout(self._timeout_req)
            self._timeout_req = None

        if self.do_stop:
            self.ioloop.stop()

        for c in self._callbacks:
            c(self)

    def callback(self, *args, **kwargs):
        """Handle result (defer is now complete"""
        self.args = args
        self.kwargs = kwargs

        self._do_done()

    def exception(self, ex_type=None, ex=None, tb=None):
        """Handle error result (defer is now complete
        
        Caller can pass in exception info, or if inside a try/except block, we
        can pull it from sys.exc_info()
        """
        if ex is None:
            ex_type, ex, tb = sys.exc_info()

        if ex and not ex_type:
            ex_type = type(ex)

        self._exception = (ex_type, ex, tb)

        self._do_done()

    def __call__(self, timeout=None):
        if not self.done and not self.ioloop:
            raise ValueError("IOLoop required")

        if not self.done:
            self.do_stop = True
            if timeout is not None:
                self._timeout_req = self.ioloop.add_timeout(time.time() + timeout, functools.partial(self.callback, error='Timeout'))

            self.ioloop.start()

        assert self.done
        return self.result

    def rethrow(self):
        if self._exception:
            ex_type, ex, tb = self._exception
            raise ex_type, ex, tb

    @property
    def error(self):
        if self._exception:
            return self._exception[1]

        return None

    @property
    def result(self):
        self.rethrow()

        if self.args and len(self.args) == 1 and not self.kwargs:
            return self.args[0]
        else:
            return self.args, self.kwargs


class ResultErrorTupleDefer(Defer):
    """Special Defer who's callback is in the form of (result, error)
    """
    @property
    def result(self):
        if len(self.args) == 2:
            return self.args[0], self.args[1]
        elif len(self.args) == 1:
            return self.args[0], None
        else:
            raise ValueError(self.args)


class ResultErrorKWDefer(Defer):
    """Special defer who's callback is in the form (result, error=ERROR)

    It will return a tuple of (result, error)
    """
    @property
    def result(self):
        if self.args:
            return self.args[0], self.kwargs.get('error')
        else:
            return None, self.kwargs.get('error')


def wait_all(all_deferred, timeout=None):

    if len(all_deferred) == 0:
        return

    ioloop = all_deferred[0].ioloop

    timeout_req = None
    if timeout:
        def timeout():
            ioloop.stop()
            raise TimeoutError()

        timeout_req = ioloop.add_timeout(time.time() + timeout, timeout)

    def callback(d):
        assert d.done
        all_deferred.remove(d)

        if len(all_deferred) == 0:
            if timeout_req:
                ioloop.remove_timeout(timeout_req)
            if ioloop: 
                ioloop.stop()

    done_defer = [d for d in all_deferred if d.done]
    [callback(d) for d in done_defer]

    for d in all_deferred:
        d.add_callback(callback)

    if all_deferred:
        ioloop.start()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
from distutils.core import setup

PACKAGES = ['dynochemy']

def get_init_val(val, packages=PACKAGES):
    pkg_init = "%s/__init__.py" % PACKAGES[0]
    value = '__%s__' % val
    fn = open(pkg_init)
    for line in fn.readlines():
        if line.startswith(value):
            return line.split('=')[1].strip().strip("'")


setup(
    name='%s' % get_init_val('title'),
    version=get_init_val('version'),
    description=get_init_val('description'),
    long_description=open('README.md').read(),
    author=get_init_val('author'),
    author_email=get_init_val('author_email'),
    url=get_init_val('url'),
    package_data={'': ['LICENSE', 'NOTICE']},
    license=get_init_val('license'),
    install_requires=['asyncdynamo>=0.2.6', 'boto>=2.6.0', 'sqlalchemy'],
    requires=['asyncdynamo', 'boto', 'sqlalchemy'],
    packages=PACKAGES
)
